//––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––– CONFIG

let SMOOTH_KEYWORD = '-s';
let FACE_KEYWORD = 'face';

let IGNORE_ROTATION_KEYWORD = '-noR';

let EYE_LEFT_KEYWORD = 'eyeL';
let EYEBOW_LEFT_KEYWORD = 'eyebowL';
let EYE_RIGHT_KEYWORD = 'eyeR';
let EYEBOW_RIGHT_KEYWORD = 'eyebowR';
let MOUTH_UPPERLIP_KEYWORD = 'mouthlipU';
let MOUTH_LOWERLIP_KEYWORD = 'mouthlipD';
let MOUTH_KEYWORD = 'mouth';
let EYELID_LEFT_UPPER_KEYWORD = 'eyelidLU';
let EYELID_RIGHT_UPPER_KEYWORD = 'eyelidRU';
let EYE_LEFT_OUTSIDE_CORNER_KEYWORD = 'eyeLOC';
let EYE_RIGHT_OUTSIDE_CORNER_KEYWORD = 'eyeROC';
let FOREHEAD_TOP_KEYWORD = 'foreheadT';

function tracked(targetRoot) {
    targetRoot.hidden = false;
}

function untracked(targetRoot) {
    targetRoot.hidden = true;
}
//–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
import { } from 'sparkar-invoke';

const Scene = require('Scene');
const FaceTracking = require('FaceTracking');
const Reactive = require('Reactive');
const Diagnostics = require('Diagnostics');

const camera = Scene.root.findFirst('Camera');
const targets = Scene.root.findByPath(`**/*${FACE_KEYWORD}*${SMOOTH_KEYWORD}*`);

let isEnable;

const FocialFeatures = {
    leftEyebrowTop: 'leftEyebrowTop',
    rightEyebrowTop: 'rightEyebrowTop',
    leftEyeCenter: 'leftEyeCenter',
    rightEyeCenter: 'rightEyeCenter',
    mouthLowerLipCenter: 'mouthLowerLipCenter',
    mouthUpperLipCenter: 'mouthUpperLipCenter',
    mouthCenter: 'mouthCenter',
    leftUpperEyelidCenter: 'leftUpperEyelidCenter',
    rightUpperEyelidCenter: 'rightUpperEyelidCenter',
    leftOutsideCorner: 'leftOutsideCorner',
    rightOutsideCorner: 'rightOutsideCorner',
    foreheadTop: 'foreheadTop',
};

//–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

function parseFocialFeatureName(name) {
    if (name.indexOf(EYE_LEFT_KEYWORD) !== -1) return FocialFeatures.leftEyeCenter;
    else if (name.indexOf(EYE_RIGHT_KEYWORD) !== -1) return FocialFeatures.rightEyeCenter;
    else if (name.indexOf(EYEBOW_LEFT_KEYWORD) !== -1) return FocialFeatures.leftEyebrowTop;
    else if (name.indexOf(EYEBOW_RIGHT_KEYWORD) !== -1) return FocialFeatures.rightEyebrowTop;
    else if (name.indexOf(MOUTH_UPPERLIP_KEYWORD) !== -1) return FocialFeatures.mouthUpperLipCenter;
    else if (name.indexOf(MOUTH_LOWERLIP_KEYWORD) !== -1) return FocialFeatures.mouthLowerLipCenter;
    else if (name.indexOf(MOUTH_KEYWORD) !== -1) return FocialFeatures.mouthCenter;
    else if (name.indexOf(EYELID_LEFT_UPPER_KEYWORD) !== -1) return FocialFeatures.leftUpperEyelidCenter;
    else if (name.indexOf(EYELID_RIGHT_UPPER_KEYWORD) !== -1) return FocialFeatures.rightUpperEyelidCenter;
    else if (name.indexOf(EYE_LEFT_OUTSIDE_CORNER_KEYWORD) !== -1) return FocialFeatures.leftOutsideCorner;
    else if (name.indexOf(EYE_RIGHT_OUTSIDE_CORNER_KEYWORD) !== -1) return FocialFeatures.rightOutsideCorner;
    else if (name.indexOf(FOREHEAD_TOP_KEYWORD) !== -1) return FocialFeatures.foreheadTop;
    else return undefined;
}

function getTrackerPosition(tracker, feature) {
    switch (feature) {
        case FocialFeatures.leftEyebrowTop:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.leftEyebrow.top).x,
                tracker.cameraTransform.applyTo(tracker.leftEyebrow.top).y,
                tracker.cameraTransform.applyTo(tracker.leftEyebrow.top).z,
            );

        case FocialFeatures.rightEyebrowTop:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.rightEyebrow.top).x,
                tracker.cameraTransform.applyTo(tracker.rightEyebrow.top).y,
                tracker.cameraTransform.applyTo(tracker.rightEyebrow.top).z,
            );

        case FocialFeatures.leftEyeCenter:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.leftEye.center).x,
                tracker.cameraTransform.applyTo(tracker.leftEye.center).y,
                tracker.cameraTransform.applyTo(tracker.leftEye.center).z,
            );

        case FocialFeatures.rightEyeCenter:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.rightEye.center).x,
                tracker.cameraTransform.applyTo(tracker.rightEye.center).y,
                tracker.cameraTransform.applyTo(tracker.rightEye.center).z,
            );

        case FocialFeatures.mouthLowerLipCenter:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.mouth.lowerLipCenter).x,
                tracker.cameraTransform.applyTo(tracker.mouth.lowerLipCenter).y,
                tracker.cameraTransform.applyTo(tracker.mouth.lowerLipCenter).z,
            );

        case FocialFeatures.mouthUpperLipCenter:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.mouth.upperLipCenter).x,
                tracker.cameraTransform.applyTo(tracker.mouth.upperLipCenter).y,
                tracker.cameraTransform.applyTo(tracker.mouth.upperLipCenter).z,
            );

        case FocialFeatures.mouthCenter:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.mouth.center).x,
                tracker.cameraTransform.applyTo(tracker.mouth.center).y,
                tracker.cameraTransform.applyTo(tracker.mouth.center).z,
            );

        case FocialFeatures.leftUpperEyelidCenter:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.leftEye.upperEyelidCenter).x,
                tracker.cameraTransform.applyTo(tracker.leftEye.upperEyelidCenter).y,
                tracker.cameraTransform.applyTo(tracker.leftEye.upperEyelidCenter).z,
            );

        case FocialFeatures.rightUpperEyelidCenter:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.rightEye.upperEyelidCenter).x,
                tracker.cameraTransform.applyTo(tracker.rightEye.upperEyelidCenter).y,
                tracker.cameraTransform.applyTo(tracker.rightEye.upperEyelidCenter).z,
            );

        case FocialFeatures.leftOutsideCorner:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.leftEye.outsideCorner).x,
                tracker.cameraTransform.applyTo(tracker.leftEye.outsideCorner).y,
                tracker.cameraTransform.applyTo(tracker.leftEye.outsideCorner).z,
            );

        case FocialFeatures.rightOutsideCorner:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.rightEye.outsideCorner).x,
                tracker.cameraTransform.applyTo(tracker.rightEye.outsideCorner).y,
                tracker.cameraTransform.applyTo(tracker.rightEye.outsideCorner).z,
            );

        case FocialFeatures.foreheadTop:
            return Reactive.pack3(
                tracker.cameraTransform.applyTo(tracker.forehead.top).x,
                tracker.cameraTransform.applyTo(tracker.forehead.top).y,
                tracker.cameraTransform.applyTo(tracker.forehead.top).z,
            );

        default:
            return Reactive.pack3(
                tracker.cameraTransform.x,
                tracker.cameraTransform.y,
                tracker.cameraTransform.z,
            );
    }
}

function smoothFollowTracker(tracker, target, smooth = 100, enableRotation = true, focialFeature) {
    return camera.then(result => {
        let trackerPosition = getTrackerPosition(tracker, focialFeature);
        trackerPosition = Reactive.pack3(
            trackerPosition.x.expSmooth(smooth),
            trackerPosition.y.expSmooth(smooth),
            trackerPosition.z.expSmooth(smooth).add(result.focalPlane.distance)
        );

        target.transform.position = trackerPosition;

        if (enableRotation) {
            const faceRotation = Reactive.pack3(
                tracker.cameraTransform.rotationX.expSmooth(smooth),
                tracker.cameraTransform.rotationY.expSmooth(smooth),
                tracker.cameraTransform.rotationZ.expSmooth(smooth)
            );
            target.transform.rotationX = faceRotation.x;
            target.transform.rotationY = faceRotation.y;
            target.transform.rotationZ = faceRotation.z;
        }
    })
}

//––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––– Method2

function enable() {
    return targets.then(all => Promise.all([
        all.map(target => {
            const indexOfFaceKeyword = target.name.indexOf(FACE_KEYWORD);
            const faceKeywordLength = FACE_KEYWORD.length;
            let faceIndex = target.name.substring(indexOfFaceKeyword + faceKeywordLength, indexOfFaceKeyword + faceKeywordLength + 1);
            faceIndex = parseInt(faceIndex);

            const indexOfSmoothKeyword = target.name.indexOf(SMOOTH_KEYWORD);
            const smoothKeywordLength = SMOOTH_KEYWORD.length;
            let smooth = target.name.substring(indexOfSmoothKeyword + smoothKeywordLength);
            smooth = parseInt(smooth);

            const tracker = FaceTracking.face(faceIndex);
            tracker.isTracked.onOn({ fireOnInitialValue: true }).subscribe(() => tracked(target));
            tracker.isTracked.onOff({ fireOnInitialValue: true }).subscribe(() => untracked(target));

            const enableRotation = target.name.indexOf(IGNORE_ROTATION_KEYWORD) === -1;

            return smoothFollowTracker(tracker, target, smooth, enableRotation, parseFocialFeatureName(target.name));
        })
    ]));
}

function disable() {
    return targets.then(all => all.forEach(target => {
        target.transform.x = target.transform.x.pinLastValue();
        target.transform.y = target.transform.y.pinLastValue();
        target.transform.z = target.transform.z.pinLastValue();

        target.transform.rotationX = target.transform.rotationX.pinLastValue();
        target.transform.rotationY = target.transform.rotationY.pinLastValue();
        target.transform.rotationZ = target.transform.rotationZ.pinLastValue();
    }));
}

module.exports = {
    get focialFeatures() {
        return FocialFeatures;
    },

    enable() {
        if (isEnable == value) return;
        isEnable = false;
        enable();
    },

    disable() {
        if (isEnable == value) return;
        isEnable = false;
        disable()
    },

    get keywords() {
        return {
            get face() { return FACE_KEYWORD },
            get smooth() { return SMOOTH_KEYWORD },
            get eye_left() { return EYE_LEFT_KEYWORD },
            get eyebow_left() { return EYEBOW_LEFT_KEYWORD },
            get eye_right() { return EYE_RIGHT_KEYWORD },
            get eyebow_right() { return EYEBOW_RIGHT_KEYWORD },
            get mouth_upperlip() { return MOUTH_UPPERLIP_KEYWORD },
            get mouth_lowerlip() { return MOUTH_LOWERLIP_KEYWORD },
            get mouth() { return MOUTH_KEYWORD },
            get eyelid_left_upper() { return EYELID_LEFT_UPPER_KEYWORD },
            get eyelid_right_upper() { return EYELID_RIGHT_UPPER_KEYWORD },
            get eye_left_outside_corner() { return EYE_LEFT_OUTSIDE_CORNER_KEYWORD },
            get eye_right_outside_corner() { return EYE_RIGHT_OUTSIDE_CORNER_KEYWORD },
            get forehead_top() { return FOREHEAD_TOP_KEYWORD },
            get ignore_rotation() { return IGNORE_ROTATION_KEYWORD },
        }
    },
}